/*
 * Gnophone: A client for the Asterisk PBX
 *
 * Copyright (C) 2000, Linux Support Services, Inc.
 *
 * Written by Mark Spencer
 *
 * Linux/UNIX version distributed under the terms of
 * the GNU General Public License
 *
 * gnophone_gui.c Graphical User Interface Code
 *
 */

#include <stdio.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>                                                   
#include <sys/signal.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include "../icons/back.xpm"
#include "../icons/forward.xpm"
#include "../icons/reload.xpm"
#include "../icons/home.xpm"
#include "../icons/stop.xpm"

#include "../icons/logo-1.xpm"
#include "../icons/logo-2.xpm"
#include "../icons/logo-3.xpm"
#include "../icons/logo-4.xpm"
#include "../icons/logo-5.xpm"

#include "gnophone.h"
#include "audio.h"
#include "telco.h"
#include "phonecore.h"
#include <string.h>
#include <sys/stat.h>
#include <gdk_imlib.h>
#include <gdk/gdkkeysyms.h>
#include <sys/time.h>
#include <netdb.h>


/*
#define ENABLE_HANDLE 
*/

/* How many times to fade */
#define MAX_FADE_LEVEL 5

/* The frequencey that we send lag requests, in ms. */
#define LAG_TIMEOUT 15000

static GtkWidget *mw;	/* Main Window */
static GtkWidget *nb;	/* Notebook */

static char extension[10] = "jpg";

static struct browser {
	void *enginepvt;
	GtkWidget *html;
	GtkWidget *frame;
	GtkWidget *spinner;
	GtkWidget *spinnerb;
	int spinnerid;
	int spinnerpos;
	GtkWidget *toolbar;
	GtkWidget *location;
	GtkWidget *back;
	GtkWidget *forward;
	GtkWidget *reload;
	GtkWidget *home;
	GtkWidget *stop;
	GtkWidget *progress;
	GtkWidget *statusbar;
	GtkWidget *detach;		/* Detach button */
	GtkWidget *outerbox;	/* Outerbox of the real browser embedded in one
	 						   of the following */
	GtkWidget *window;		/* Stand alone window */
	GtkWidget *box;			/* Box in the Notebook */
	struct html_engine *engine;
} *pbrowser;				/* Primary browser window */

#define DONT_DEFINE_BROWSER
#include "htmlengine.h"

static struct peer {
	GtkWidget *image;
	GtkWidget *button;
	GtkWidget *level;
	GtkWidget *volume;
	GtkWidget *frame;
	GtkWidget *frame2;
	GtkWidget *label;
	GtkWidget *outerbox;
	GtkWidget *box;
	GtkWidget *status;
	GtkWidget *hangup;
	GtkWidget *answer;
	GtkWidget *hold;
	GtkWidget *conf;
	GtkWidget *direct;
	GtkWidget *sendimage;
	GtkWidget *lagmeter;
	
	/* Image to fade from */
	GdkImlibImage *fade;
	GdkImlibImage *orig;

	/* How much we've faded */
	int fadelevel;
	int holdid;
	int lagfd;
	float history;
	float blah;
	int laglock;

	struct sockaddr_in addr;

	int mute;
	int conference;

	int xoff;
	int yoff;
	int moving;

	/* Phonecore ID */
	int pcid;
	struct peer *next;
} *self;

static int regrefresh = 60;

struct html_engine *engine=NULL, *engines=NULL;

static struct peer *peers;
static struct peer *current_peer;
static struct peer *most_recent_answer;
static int iaxloading = 0;
static struct peer *urlpeer;

static GtkWidget *fixed;
static GtkWidget *fixedstatus;
static GtkWidget *menubar;
static GtkAccelGroup *gag;
static GtkWidget *callbttn;
static GtkWidget *callwin;
static GtkWidget *calladdr;
static GtkWidget *cancelbttn;

static GdkPixmap *spinner_pm[5];
static GdkBitmap *spinner_bm[5];

static int registry = -1;
int regid = -1;
int tbd_pcid = -1;

static void select_peer(struct peer *p);
void show_call_window(int, int);
static int send_lag_request(gpointer data);

int html_register_engine(struct html_engine *e)
{
	static struct html_engine *last = NULL;
	printf("Registering %s\n", e->name);
	e->next = NULL;

	if (last)
		last->next = e;
	else
		engines = e;

	last = e;
	return 0;
}

int html_select_engine(struct html_engine *e)
{
	char *s=NULL;
	if (engine) {
		engine = e;
		if (pbrowser) {
			if (pbrowser && pbrowser->location)
				s = gtk_entry_get_text(GTK_ENTRY(pbrowser->location));
			/* Destroy the old widget */
			gtk_widget_destroy(pbrowser->html);
			pbrowser->html = engine->new_engine(pbrowser);
			/* Add it to the frame */
			gtk_container_add(GTK_CONTAINER(pbrowser->frame), pbrowser->html);
			/* Show it */
			gtk_widget_show_all(pbrowser->html);
			pbrowser->engine = engine;
			if (s && strlen(s))
				pbrowser->engine->loadurl(pbrowser, s);
		}
	} else
		engine = e;
	return 0;
}

static void init_spinner()
{
	spinner_pm[0] = gdk_pixmap_create_from_xpm_d(mw->window, spinner_bm, NULL, logo_1_xpm);
	spinner_pm[1] = gdk_pixmap_create_from_xpm_d(mw->window, spinner_bm + 1, NULL, logo_2_xpm);
	spinner_pm[2] = gdk_pixmap_create_from_xpm_d(mw->window, spinner_bm + 2, NULL, logo_3_xpm);
	spinner_pm[3] = gdk_pixmap_create_from_xpm_d(mw->window, spinner_bm + 3, NULL, logo_4_xpm);
	spinner_pm[4] = gdk_pixmap_create_from_xpm_d(mw->window, spinner_bm + 4, NULL, logo_5_xpm);
}

static int spinner_cb(void *v)
{
	struct browser *b = v;
	b->spinnerpos++;
	if (b->spinnerpos > 4)
		b->spinnerpos = 0;
	gtk_pixmap_set(GTK_PIXMAP(b->spinner), spinner_pm[b->spinnerpos], spinner_bm[b->spinnerpos]);
	return TRUE;
}

static int stop_spinner(struct browser *b)
{
	b->spinnerpos = 0;
	gtk_timeout_remove(b->spinnerid);
	gtk_pixmap_set(GTK_PIXMAP(b->spinner), spinner_pm[0], spinner_bm[0]);
	b->spinnerpos =0;
	return 0;
}

static int start_spinner(struct browser *b)
{
	b->spinnerpos = 0;
	b->spinnerid = gtk_timeout_add(200, spinner_cb, b);	
	return 0;
}

void html_display_new_location(struct browser *b, char *s)
{
	gtk_entry_set_text(GTK_ENTRY(b->location), s);
}

void html_display_new_title(struct browser *b, char *s)
{
	char buf[1024];
	snprintf(buf, sizeof(buf),"Gnophone: %s", s);
	gtk_window_set_title(GTK_WINDOW(mw), buf);
}	

void html_display_load_started(struct browser *b)
{
	gtk_widget_set_sensitive(b->stop, TRUE);	
	gtk_progress_set_format_string(GTK_PROGRESS(b->progress), "%p%%");
	gtk_progress_set_show_text(GTK_PROGRESS(b->progress), TRUE);
	gtk_progress_configure(GTK_PROGRESS(b->progress), 0, 0, 1);
	start_spinner(b);
}

static int do_registry(void *v)
{
	regid = -1;
	iax_reg();
	return FALSE;
}

static int schedule_registry()
{
	if (regid > -1)
		gtk_timeout_remove(regid);
#if 0
	printf("Registering in %d ms\n", regrefresh * 5000/6);
#endif	
	regid = gtk_timeout_add(regrefresh * 5000 / 6, do_registry, NULL);
	return 0;
}

static void update_toolbar(struct browser *b)
{
	gtk_widget_set_sensitive(b->back,
			b->engine->can_go_back(b));
	gtk_widget_set_sensitive(b->forward,
			b->engine->can_go_forward(b));
}

void html_display_load_finished(struct browser *b)
{
	gtk_widget_set_sensitive(b->stop, FALSE);	
	if (iaxloading && urlpeer && (urlpeer->pcid > -1))
		pc_load_complete(SOURCE_GUI, urlpeer->pcid);
	iaxloading = 0;
	stop_spinner(b);
	update_toolbar(b);
}

void html_setstatus(struct browser *b, char *text)
{
	if (b) {
		gtk_statusbar_pop(GTK_STATUSBAR(b->statusbar), 1);
		gtk_statusbar_push(GTK_STATUSBAR(b->statusbar), 1, text);
	}
}

static void setpeerstatus(struct peer *p, char *text)
{
	GtkWidget *statusbar;
	if (p) 
		statusbar = p->status;
	else 
		statusbar = fixedstatus;
	gtk_statusbar_pop(GTK_STATUSBAR(statusbar), 1);
	gtk_statusbar_push(GTK_STATUSBAR(statusbar), 1, text);
}

void html_update_progress(struct browser *b, int cur, int max)
{
	if (max > 1048576) {
		gtk_progress_set_format_string(GTK_PROGRESS(b->progress), "%p%% of %v megs");
		cur /= 1048576;
		max /= 1048576;
	} else if (max > 1024) {
		gtk_progress_set_format_string(GTK_PROGRESS(b->progress), "%p%% of %v kbytes");
		cur /= 1024;
		max /= 1024;
	} else {
		gtk_progress_set_format_string(GTK_PROGRESS(b->progress), "%p%% of %v bytes");
	}
	/* Fake it if we don't know how big the document is */
	if (max < 0)
		max = cur * 20;

	if (max)
		gtk_progress_configure(GTK_PROGRESS(b->progress), cur, 0, max);
}

void html_display_link(struct browser *b, char *s)
{
	html_setstatus(b, s);
}

static void back_cb(GtkWidget *w, struct browser *b)
{
	b->engine->go_back(b);
}

static void forward_cb(GtkWidget *w, struct browser *b)
{
	b->engine->go_forward(b);
}

static void reload_cb(GtkWidget *w, struct browser *b)
{
	b->engine->reload(b, 1);
}

static void home_cb(GtkWidget *w, struct browser *b)
{
	b->engine->loadurl(b, home_page);
}

static void stop_cb(GtkWidget *w, struct browser *b)
{
	b->engine->stop(b);
}

static void entry_cb(GtkWidget *w, struct browser *b)
{
	char *text = gtk_entry_get_text(GTK_ENTRY(b->location));	
	if (text) {
		interpret_call(text);
#if 0
		if(strlen(text) >= 6 && text[5] == '/') {
			text[5] = 0;
			if(!strcasecmp(text, "tel:/") || !strcasecmp(text, "iax:/")) {
				char *num = &text[6];
				puts("calling.");
				launch_call(num, num);
			}
		}
		else b->engine->loadurl(b, text);
#endif		
	}
}

int _start_load_url(struct browser *b, char *url)
{
	gtk_notebook_set_page(GTK_NOTEBOOK(nb), 0);
	b->engine->loadurl(b, url);
	return 0;
}

static GtkWidget *mk_pixmap(struct browser *b, char **data)
{
	GdkPixmap *pm;
	GdkBitmap *bm;
	pm = gdk_pixmap_create_from_xpm_d(mw->window, &bm, NULL, data);
	return gtk_pixmap_new(pm, bm);
}

static GtkWidget *selfb;
static GtkWidget *gwin;

#ifdef ENABLE_HANDLE
static void detached_cb(GtkHandleBox *h, GtkWidget *w, struct browser *b)
{
	gdk_window_resize(GTK_WIDGET(h)->window, b->outerbox->allocation.width,
											 b->outerbox->allocation.height);
	gtk_handle_box_set_shadow_type(h, GTK_SHADOW_IN);									
	NOTE;
}
static void attached_cb(GtkHandleBox *h, GtkWidget *w, struct browser *b)
{
	gtk_handle_box_set_shadow_type(h, GTK_SHADOW_NONE);									
	NOTE;
}
#endif

static void dialdigit(GtkWidget *w, char *digit)
{
	send_digit(*digit);
	if (current_peer && (current_peer->pcid > -1)) {
#if 1
		printf("Sending digit %c\n", *digit);
#endif
		pc_send_dtmf(SOURCE_GUI, current_peer->pcid, *digit);
	} else {
		if (!callwin)
			show_call_window(0, 0);
		gtk_entry_append_text(GTK_ENTRY(calladdr), digit);
	}
}

static GtkWidget *build_dialpad()
{
	GtkWidget *table;
	GtkWidget *button;
	GtkWidget *frame;
	static char *labels[] = 
	{ "1", "2", "3", "4", "5", "6", "7", "8", "9", "*", "0", "#" };
	int x;
	frame = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_OUT);
	table = gtk_table_new(4, 3, TRUE);
	gtk_container_set_border_width(GTK_CONTAINER(table), 10);
	for (x=0;x<12;x++) {
		button = gtk_button_new_with_label(labels[x]);
		gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(dialdigit), labels[x]);
		gtk_widget_set_usize(button, 35,25);
		if ((labels[x][0] >='0') && (labels[x][0] <= '9')) {
			gtk_widget_add_accelerator(button, "clicked", gag, GDK_KP_0+atoi(&labels[x][0]), 0, 0);
			gtk_widget_add_accelerator(button, "clicked", gag, GDK_0+atoi(&labels[x][0]), 0, 0);
		} else {
			gtk_widget_add_accelerator(button, "clicked", gag, labels[x][0], GDK_SHIFT_MASK, 0);
		}
		gtk_object_set_user_data(GTK_OBJECT(button), labels[x]);
		gtk_table_attach(GTK_TABLE(table), button,
						 x % 3, x % 3 + 1,
						 x / 3, x / 3 + 1,
						 GTK_FILL, GTK_FILL,
						 5, 5);
	}
	gtk_container_add(GTK_CONTAINER(frame), table);
	return frame;
}


static void about(void)
{
	printf("About!\n");
}

static struct browser *build_browser(void)
{
	struct browser *b;
	GtkWidget *hbox;
	GtkWidget *label;

	if (!engine)
		return NULL;

	/* Initialize spinner pixmaps */
	init_spinner();

	/* Create a browser structure */
	b = g_new0(struct browser, 1);
	
	b->engine = engine;
	
	/* Create an browser widget */
	b->html = b->engine->new_engine(b);
	
	
	/* Put it in a frame */
	b->frame = gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(b->frame), b->html);

	/* Create a toolbar */
	b->toolbar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_BOTH);
	gtk_toolbar_set_space_size(GTK_TOOLBAR(b->toolbar), 10);

	/* Create the main buttons with space between all of them */
	b->back = gtk_toolbar_append_item(GTK_TOOLBAR(b->toolbar), "Back", 
										"Return to previous page", "back", mk_pixmap(b, back), back_cb, b);
	gtk_toolbar_append_space(GTK_TOOLBAR(b->toolbar));
	b->forward = gtk_toolbar_append_item(GTK_TOOLBAR(b->toolbar), "Forward", 
										"Go forward", "forward", mk_pixmap(b, forward), forward_cb, b);
	gtk_toolbar_append_space(GTK_TOOLBAR(b->toolbar));
	b->reload = gtk_toolbar_append_item(GTK_TOOLBAR(b->toolbar), "Reload", 
										"Reload current page", "reload", mk_pixmap(b, reload), reload_cb, b);
	gtk_toolbar_append_space(GTK_TOOLBAR(b->toolbar));
	gtk_toolbar_append_space(GTK_TOOLBAR(b->toolbar));
	b->home = gtk_toolbar_append_item(GTK_TOOLBAR(b->toolbar), "Home", 
										"Go To Home Page", "home", mk_pixmap(b, home), home_cb, b);
	gtk_toolbar_append_space(GTK_TOOLBAR(b->toolbar));
	gtk_toolbar_append_space(GTK_TOOLBAR(b->toolbar));
	b->stop = gtk_toolbar_append_item(GTK_TOOLBAR(b->toolbar), "Stop", 
										"Stop loading page", "stop", mk_pixmap(b, stop), stop_cb, b);

	gtk_toolbar_append_space(GTK_TOOLBAR(b->toolbar));
	gtk_toolbar_append_space(GTK_TOOLBAR(b->toolbar));
	b->spinner = gtk_pixmap_new(spinner_pm[0], spinner_bm[0]);
#if 0
	b->spinnerb = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(b->spinnerb), b->spinner);
	gtk_toolbar_append_element(GTK_TOOLBAR(b->toolbar), GTK_TOOLBAR_CHILD_WIDGET, b->spinnerb, "", "Fancy Spinner Thing", "spinner", NULL, 
								GTK_SIGNAL_FUNC(about), b);

#else
	b->spinnerb = gtk_button_new();
	gtk_toolbar_append_element(GTK_TOOLBAR(b->toolbar), GTK_TOOLBAR_CHILD_WIDGET, b->spinner, "", "Fancy Spinner Thing", "spinner", NULL, 
								GTK_SIGNAL_FUNC(about), b);
#endif
	/* Create a horizontal box and put the locator in it */
	b->location = gtk_entry_new();
	gtk_signal_connect(GTK_OBJECT(b->location), "activate", GTK_SIGNAL_FUNC(entry_cb), b);
	
	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new("Location:");
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX(hbox), b->location, TRUE, TRUE, 0);


	/* Pack stuff in the box */
	b->outerbox = gtk_vbox_new(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(b->outerbox), 5);
	gtk_box_pack_start(GTK_BOX(b->outerbox), b->toolbar, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX(b->outerbox), hbox, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX(b->outerbox), b->frame, TRUE, TRUE, 5);

	/* And we should have a status bar at the bottom */
	b->statusbar = gtk_statusbar_new();
	html_setstatus(b, "Ready.");

	/* Make an h-box for the status bar */
	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), b->statusbar, TRUE, TRUE, 0);
	
	/* And a progress bar */
	b->progress = gtk_progress_bar_new();
	gtk_box_pack_start(GTK_BOX(hbox), b->progress, FALSE, FALSE, 0);
	
	gtk_box_pack_start(GTK_BOX(b->outerbox), hbox, FALSE, FALSE, 0);

	/* Make sure the window is NULL */
	b->window = NULL;

	gtk_widget_set_sensitive(b->stop, FALSE);	
	
	/* Create a box container */
#ifdef ENABLE_HANDLE
	b->box = gtk_handle_box_new();
	gtk_handle_box_set_shadow_type(GTK_HANDLE_BOX(b->box), GTK_SHADOW_NONE);
	gtk_signal_connect(GTK_OBJECT(b->box), "child-detached", detached_cb, b);
	gtk_signal_connect(GTK_OBJECT(b->box), "child-attached", attached_cb, b);
#else
	b->box = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(b->box), GTK_SHADOW_NONE);
#endif
	gtk_container_add(GTK_CONTAINER(b->box), b->outerbox);
	
	/* Update the toolbar */
	update_toolbar(b);	
		
	return b;
}

int start_load_url(char *url)
{
	if (pbrowser)
		return _start_load_url(pbrowser, url);
	return 0;
}
	

int init_gphx_subsystem(int *argc, char **argv[])
{
	if (gtk_init_check(argc, argv)) {
		/* GUI mode success */
		gtk_init(argc, argv);
		gdk_imlib_init();
		gtk_widget_push_visual(gdk_imlib_get_visual());
		gtk_widget_push_colormap(gdk_imlib_get_colormap());
	} else {
		/* No GUI? */
		fprintf(stderr, "Unable to initialize X-windows\n");
		return -1;
	}
	return 0;
}

int quit(GtkWidget *widget, void *data)
{
	stop_camera();
	signal(SIGCHLD, NULL);
	gtk_main_quit();
	return TRUE;
}

int stop_image(void *wv)
{
	GtkWidget *w = wv;
	gdk_window_set_back_pixmap(w->window, NULL, 0);
	return 0;
}

int mk_image(void *wv, char *filename)
{
	GtkWidget *w = wv;
	GdkImlibImage *im;
	im = gdk_imlib_load_image(filename);
	gdk_imlib_apply_image(im, w->window);
	gdk_imlib_kill_image(im);
	return 0;
}

char *user_image_filename()
{
	static char buf[1024];

	if (getenv("HOME")) {
		snprintf(buf, sizeof(buf), "%s/.gnophone", getenv("HOME"));
		mkdir(buf, 0700);

		snprintf(buf, sizeof(buf), "%s/.gnophone/my_picture.%s", getenv("HOME"), extension);
		return buf;
	}
	return NULL;		
}

void image_callback(void)
{
	int wasusingcam = use_cam();

	stop_camera();
	gtk_widget_set_sensitive(mw, FALSE);
	show_video_control();
	gtk_widget_set_sensitive(mw, TRUE);
	if (!wasusingcam)
		stop_image(gwin);
	if (use_cam()) {
		set_cam_output(gwin);
		start_camera();
	} else {
		mk_image(gwin, get_image_filename());
	}
}

static void audio_prefs(void)
{
	gtk_widget_set_sensitive(mw, FALSE);
	show_audio_control();
	gtk_widget_set_sensitive(mw, TRUE);
}

void telco_prefs(void)
{
	gtk_widget_set_sensitive(mw, FALSE);
	telco_show_gui();
	gtk_widget_set_sensitive(mw, TRUE);
}

char *get_image_filename()
{
	struct stat s;

	if (!stat(user_image_filename(), &s) && (video_options & OPT_VIDEO_SEND_IMAGE))
		return user_image_filename();

	return base_filename(DEFAULT_IMAGE);
}

static GtkWidget *build_self_win(void)
{
	GtkWidget *f;
	/* This is a view of ourselves */
	selfb = gtk_button_new();
	gtk_signal_connect(GTK_OBJECT(selfb), "clicked", image_callback, NULL);
	gwin = gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(gwin), VIDEO_WIDTH,VIDEO_HEIGHT);
	f = gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(f), gwin);
	gtk_container_add(GTK_CONTAINER(selfb), f);
	return selfb;
	
}

static void generic_callback(GtkWidget *w)
{
	printf("Generic Callback!\n");
}

static void send_new_image(GtkWidget *w, gpointer *data)
{
	struct peer *p = (struct peer *)data;
	GdkImlibImage *snapshot;
	char *newname = base_filename(TEMP_IMAGE);

	if (gwin) {
		/* Get a snapshot of the camera */
		snapshot = gdk_imlib_create_image_from_drawable(gwin->window, NULL, 0, 0, VIDEO_WIDTH, VIDEO_HEIGHT);

		/* Save it to disk */
		gdk_imlib_save_image(snapshot, newname, NULL);

		/* And destroy the evidence ;-) */
		gdk_imlib_destroy_image(snapshot);

		/*  Now, we should tell phonecore to send it. */
		printf("Sending image to PCID: %d\n", p->pcid);
		send_my_image(p->pcid, newname);	
	}
}

static GtkWidget *build_generic_win(struct peer *p)
{
	GtkWidget *f;
	GtkWidget *y;
	GtkWidget *b;
	GtkWidget *hbox;
	GtkWidget *l;
	
	y = gtk_vbox_new(FALSE, 0);
	gtk_box_set_spacing(GTK_BOX(y), 5);

	/* This is a view of ourselves */
	p->button = gtk_button_new();
	gtk_signal_connect(GTK_OBJECT(selfb), "clicked", generic_callback, NULL);
	p->image = gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(p->image), VIDEO_WIDTH, VIDEO_HEIGHT);
	f = gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(f), p->image);
	gtk_container_add(GTK_CONTAINER(p->button), f);

	hbox = gtk_hbox_new(FALSE, 0);
	l = gtk_label_new(NULL);
	gtk_box_pack_start(GTK_BOX(hbox), l, TRUE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(p->button), FALSE, FALSE, 0);
	l = gtk_label_new(NULL);
	gtk_box_pack_start(GTK_BOX(hbox), l, TRUE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(y), GTK_WIDGET(hbox), FALSE, FALSE, 0);

	p->sendimage = gtk_button_new_with_label("Send Image");
	gtk_signal_connect(GTK_OBJECT(p->sendimage), "clicked", send_new_image, p);
	gtk_box_pack_start(GTK_BOX(y), GTK_WIDGET(p->sendimage), FALSE, FALSE, 0);

	return y;
	
}

static GtkWidget *callanswer;

/* No more than 4 concurrent IAX sessions */
#define MAX_SESSIONS 4
static int sessions = 0;

static int real_destroy_peer_timeout(void *vp)
{
	struct peer *p = vp;
	if (p->fade)
		gdk_imlib_kill_image(p->fade);
	if (p->orig)
		gdk_imlib_kill_image(p->orig);
	gtk_widget_destroy(p->box);
	free(p);
	if (!peers)
		gtk_notebook_set_page(GTK_NOTEBOOK(nb), 0);
	return FALSE;
}

static int peer_fade_some(void *vp)
{
	struct peer *p = vp;
	int x;
	unsigned char *rgbin;
	unsigned char *rgbout;
	float f;
#if 0
	while(gtk_events_pending()) gtk_main_iteration_do(FALSE);
	gdk_imlib_create_image_from_drawable(fixed->window, 
			NULL, p->box->allocation.x, p->box->allocation.y, p->box->allocation.width, p->box->allocation.height);												
	gtk_widget_show(p->box);
	while(gtk_events_pending()) gtk_main_iteration_do(FALSE);
#endif
	/* Take a weighted average of the background and ourselves */
	rgbin = p->orig->rgb_data;
	rgbout = p->fade->rgb_data;
	for (x=0;x<p->orig->rgb_width * p->orig->rgb_height * 3; x++) {
		/* Use a portion of gray */
		switch(x % 3) {
		case 0:	/* Red */
			f = 208.0;
			break;
		case 1: /* Green */
			f = 212.0;
			break;
		case 2: /* Blue */
			f = 208.0;
			break;
		}
		f *= (float)(p->fadelevel) / (float)(MAX_FADE_LEVEL);
		f += (float)p->orig->rgb_data[x] * (float)(MAX_FADE_LEVEL - p->fadelevel) / ((float)(MAX_FADE_LEVEL));
		rgbout[x] =  f;
	}

	gdk_imlib_changed_image(p->fade);
	gdk_imlib_apply_image(p->fade, p->frame2->window);
	if (p->fadelevel++ > MAX_FADE_LEVEL - 1) {
		gtk_timeout_add(100, real_destroy_peer_timeout, p);
		return FALSE;
	}
	return TRUE;
}

static int drawing_configure(GtkWidget *w, GdkEventConfigure *c, void *vp)
{
	struct peer *p = vp;
	gdk_imlib_apply_image(p->fade, p->frame2->window);
	return TRUE;
}

static int destroy_peer_timeout(void *vp)
{
	struct peer *p = vp;
	int x,y, w, h;
	/* Get a picture of what we look like now */
	p->fade = gdk_imlib_create_image_from_drawable(p->frame2->window, 
			NULL, 0, 0, p->frame2->allocation.width, p->frame2->allocation.height);
	if (!p->fade) {
		fprintf(stderr, "Nobody looking.  Destroying us now\n");
		/* Nobody is looking.  Just go ahead and destroy us completely */
		real_destroy_peer_timeout(p);
		return FALSE;
	}
	p->orig = gdk_imlib_clone_image(p->fade);

	/* Save original position */
	w = p->box->allocation.width;
	h = p->box->allocation.height;
	x = p->box->allocation.x;
	y = p->box->allocation.y;

	gtk_widget_destroy(p->frame2);
	p->frame2 = gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(p->frame2), w, h);
	gtk_container_add(GTK_CONTAINER(p->box), p->frame2);

#if 0
	gtk_fixed_put(GTK_FIXED(fixed), p->box, x, y);
#endif

	gtk_signal_connect(GTK_OBJECT(p->frame2), "configure-event", GTK_SIGNAL_FUNC(drawing_configure), p);	
	gtk_widget_show_all(p->box);
	
	gtk_timeout_add(200, peer_fade_some, p);
	return FALSE;
}

static void remove_peer(struct peer *p)
{
	struct peer *cur, *prev = NULL;
	cur = peers;
	while(cur) {
		if (cur == p) {
			if (prev)
				prev->next = p->next;
			else
				peers = p->next;
		}
		prev = cur;
		cur = cur->next;
	}
}			

static void handle_hangup(struct peer *p, char *reason)
{
	char buf[80];
	/* Got a hangup, destroy the peer */
	
	/* Make sure we don't have any timeouts going */
	while (p->laglock) {
	}

	/* Remove the timeout */
	if (p->lagfd > -1) {
		gtk_timeout_remove(p->lagfd);
		p->lagfd = -1;
	}

	
	audio_shutup();
	audio_deactivate();
	remove_peer(p);
	if (most_recent_answer == p)
		most_recent_answer = NULL;
	if (p == current_peer) {
		current_peer = peers;
	} else
		printf("Hm... it's not me, it's %p\n", current_peer);
	p->pcid = -1;
	if (urlpeer == p)
		urlpeer = NULL;
	p->pcid = -1;
	 if (p->holdid > 0)
		gtk_timeout_remove(p->holdid);
	p->holdid = -1;
	setpeerstatus(p, reason);
	gtk_widget_set_sensitive(p->frame2, FALSE);
	gtk_timeout_add(5000, destroy_peer_timeout, p);
	sessions--;
	fprintf(stderr, "Sessions=%d?\n", sessions);
	if (sessions < 0) {
		fprintf(stderr, "Sessions is now %d?\n", sessions);
	}
	if (sessions > 0) {
		snprintf(buf, sizeof(buf), "%d active call%s", sessions, (sessions != 1 ? "s" : ""));
		setpeerstatus(NULL, buf);
	} else
		setpeerstatus(NULL, "Call disconnected.");
}

static void peerhangup_cb(GtkWidget *w, struct peer *p)
{


	pc_hangup(SOURCE_GUI, p->pcid, "Goodbye.");
	handle_hangup(p, "Call disconnected.");
}

static int holdblink(void *data);
static int holdblinkoff(void *data)
{
	struct peer *p = data;
	gtk_widget_set_state(GTK_WIDGET(p->hold), GTK_STATE_NORMAL);
	p->holdid = gtk_timeout_add(100, holdblink, p);
	return 0;
}
static int holdblink(void *data)
{
	struct peer *p = data;
	gtk_widget_set_state(GTK_WIDGET(p->hold), GTK_STATE_PRELIGHT);
	p->holdid = gtk_timeout_add(400, holdblinkoff, p);
	return 0;
}

static void fake_set_active(GtkWidget *w, int val)
{
	if (val)
		val = 1;
	else
		val = 0;
	GTK_TOGGLE_BUTTON(w)->active = val;
	if (val)
		gtk_widget_set_state(w, GTK_STATE_ACTIVE);
	else
		gtk_widget_set_state(w, GTK_STATE_NORMAL);
}

static void holdon(struct peer *p)
{
	fake_set_active(p->hold, TRUE);
	fake_set_active(p->direct, FALSE);
	fake_set_active(p->conf, FALSE);
	select_peer(NULL);
	p->holdid = gtk_timeout_add(1000, holdblink, p);
	if (p->conference)
		pc_unconference(SOURCE_GUI, p->pcid);
}

static void holdoff(struct peer *p)
{
	fake_set_active(p->hold, FALSE);
	if (p->holdid > -1)
		gtk_timeout_remove(p->holdid);
	p->holdid = -1;
	gtk_widget_set_state(GTK_WIDGET(p->hold), GTK_STATE_NORMAL);
	if (p->conference) {
		pc_conference(SOURCE_GUI, p->pcid);
		fake_set_active(p->direct, FALSE);
		fake_set_active(p->conf, TRUE);
	} else {
		pc_unconference(SOURCE_GUI, p->pcid);
		select_peer(p);
		fake_set_active(p->direct, TRUE);
		fake_set_active(p->conf, FALSE);
	}
}

static void peerhold_cb(GtkWidget *w, struct peer *p)
{
	if (GTK_TOGGLE_BUTTON(w)->active) {
		holdon(p);
	} else {
		holdoff(p);
	}
}

static void conference_cb(GtkWidget *w, struct peer *p)
{
	printf("confcb!\n");
	if (GTK_TOGGLE_BUTTON(w)->active) {
		p->conference = 1;
		holdoff(p);
	} else {
		p->conference = 0;
		pc_unconference(SOURCE_GUI, p->pcid);
		holdon(p);
	}
}

static void direct_cb(GtkWidget *w, struct peer *p)
{
	if (GTK_TOGGLE_BUTTON(w)->active) {
		p->conference = 0;
		holdoff(p);
	} else {
		holdon(p);
	}
}

static void hold_peer(struct peer *p)
{
	if (!GTK_TOGGLE_BUTTON(p->hold)->active) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(p->hold), TRUE);
	}
}

static void select_peer(struct peer *p)
{
	if (current_peer && (current_peer != p) && (!current_peer->conference))
		hold_peer(current_peer);
	current_peer = p;
	if (current_peer)
		pc_select(SOURCE_GUI, current_peer->pcid);
	else
		pc_select(SOURCE_GUI, -1);
}

static void do_answer(struct peer *p)
{
	char buf[80];

	pc_answer(SOURCE_GUI, p->pcid);
	p->mute = 0;
	gtk_widget_set_sensitive(p->answer, FALSE);
	setpeerstatus(p, "Connected.");
	snprintf(buf, sizeof(buf), "%d active call%s", sessions, (sessions != 1 ? "s" : ""));
	setpeerstatus(NULL, buf);
	select_peer(p);
	if (most_recent_answer == p)	
		most_recent_answer = NULL;

	audio_shutup();

	if (video_options & OPT_VIDEO_SEND_IMAGE)
		if (user_image_filename())
			send_my_image(p->pcid, user_image_filename());	
}

void send_my_image(int callno, char *filename)
{
	pc_event e;
	int fd;
	char data[16384];
	int size = 0;
	struct stat s;

	/* Check to see if the file exists */
	if (stat(filename, &s))
		return;

	fd = open(filename, O_RDONLY);
	if (fd < 0)
		return;

	size = read(fd, data, sizeof(e.e.image.data));

	close(fd);

	if (size < 0) 
		return;

	e.event = PC_EVENT_IMAGE;
	e.len = sizeof(pc_event);
	e.callno = callno;
	e.e.image.datalen = size;

	if (size > sizeof(e.e.image.data))
		size = sizeof(e.e.image.data);

	memcpy(e.e.image.data, data, size);

	pc_write_event(SOURCE_GUI, &e);
}

void answer(void)
{
	if (most_recent_answer)
		do_answer(most_recent_answer);
}

static void peeranswer_cb(GtkWidget *w, struct peer *p)
{
	do_answer(p);
}

static void callwin_destroy(void)
{
	gtk_grab_remove(callwin);
	callwin = NULL;
	gtk_widget_set_sensitive(mw, TRUE);
}

static void hide_call_window(void)
{
	if (callwin)
		gtk_widget_destroy(callwin);
}

static void place_call()
{
	int uc = use_cam();
	char *s = strdup(gtk_entry_get_text(GTK_ENTRY(calladdr)));
	gtk_widget_destroy(callwin);

puts("got entry.");

	/* If using the camera, stop it briefly */
	if (uc)
		stop_camera();

	while(gtk_events_pending())
		gtk_main_iteration_do(FALSE);

	/* and start it back up */
	if (uc)
		start_camera();

	if (strlen(s)) {

		/* We don't want it to default to http:// as the protocol. This was 
		* changed as per Mark's request. */

		if (!strstr(s, "://")) {
			char *s2 = (char *)malloc((strlen(s) + 7) * sizeof(char));

			if (!is_phone_number(s))
			{

				/* Ok, we don't have a phone number. Is it 
				* a host name or an extension */

				if (strstr(s, ".")) {

					/* Looks like it's a host name. */
					sprintf(s2, "iax://%s", s);

					free(s);
					s = strdup(s2);
				}
			}

			free(s2);
		}

		if (interpret_call(s)) {
			setpeerstatus(NULL, "Invalid call address");
			html_setstatus(pbrowser, "Invalid call address");
		}
	}

	free(s);
}

void hangup()
{
	if (current_peer) {
		if (current_peer->pcid >= 0) {
			pc_hangup(SOURCE_GUI, current_peer->pcid, "Goodbye.");
			handle_hangup(current_peer, "Call disconnected.");
		}

	} else if (callwin) {
		gtk_button_clicked(GTK_BUTTON(cancelbttn));
	}
	if (self && self->level) 
		gtk_progress_bar_update(GTK_PROGRESS_BAR(self->level), 0.0);
}

void show_call_window(int block, int disable)
{
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *label;

	/* Make our boxes */	
	vbox = gtk_vbox_new(FALSE, 5);
	hbox = gtk_hbox_new(FALSE, 5);

	/* Make a label */
	label = gtk_label_new("Please enter an Internet URL, Internet host \nsupporting IAX, or a simple phone number");
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
	
	/* Make a spot for them to put in who they want to call */
	calladdr = gtk_entry_new();
	gtk_widget_set_usize(calladdr, 300, 0);
	gtk_signal_connect(GTK_OBJECT(calladdr), "activate", GTK_SIGNAL_FUNC(place_call), NULL);
	
	/* Make our buttons */
	callbttn = gtk_button_new_with_label("Call");
	gtk_widget_add_accelerator(callbttn, "clicked", gag, GDK_Return, 0, 0);
	gtk_widget_set_usize(callbttn, 80, 30);
	gtk_signal_connect(GTK_OBJECT(callbttn), "clicked", GTK_SIGNAL_FUNC(place_call), NULL);
	cancelbttn = gtk_button_new_with_label("Cancel");
	gtk_widget_set_usize(cancelbttn, 80, 30);
	gtk_signal_connect(GTK_OBJECT(cancelbttn), "clicked", GTK_SIGNAL_FUNC(hide_call_window), NULL);

	/* Pack them into the boxes */
	gtk_box_pack_start(GTK_BOX(hbox), callbttn, FALSE, FALSE, 40);
	gtk_box_pack_end(GTK_BOX(hbox), cancelbttn, FALSE, FALSE, 40);
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 10);
	gtk_box_pack_start(GTK_BOX(vbox), calladdr, FALSE, FALSE, 10);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 10);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);
	
	if (disable)
		gtk_widget_set_sensitive(mw, FALSE);
	
	callwin = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_container_add(GTK_CONTAINER(callwin), vbox);
	gtk_window_set_title(GTK_WINDOW(callwin), "Gnophone: Place a call");
	gtk_widget_show_all(callwin);

	if (disable)
		gtk_grab_add(callwin);

	gtk_signal_connect(GTK_OBJECT(callwin), "destroy", GTK_SIGNAL_FUNC(callwin_destroy), NULL);
	gtk_widget_grab_focus(calladdr);

	if (block)
		while(callwin && !gtk_main_iteration_do(TRUE));
}

void digit_ready(char digit, int startok, int useurl)
{
	char s[2];

	if (useurl) {
		if (urlpeer && (urlpeer->pcid > -1))
			pc_send_dtmf(SOURCE_GUI, urlpeer->pcid, digit);
		return;
	}

	if (current_peer && (current_peer->pcid > -1)) {
		/* We're in a call, so send a digit */
		pc_send_dtmf(SOURCE_GUI, current_peer->pcid, digit);
	} else if (startok) {
		/* We're not in a call, so send our digit to the "call" window, bringing
		   it up if it's not already */
		if (!callwin)
			show_call_window(0, 0);
		s[0] = digit;
		s[1] = '\0';
		if ((digit >= '0') && (digit <= '9')) 
			gtk_entry_append_text(GTK_ENTRY(calladdr), s);
		else {	
			if (strlen(gtk_entry_get_text(GTK_ENTRY(calladdr)))) {
				/* If this isn't the first character of the string, like
				   "*67", then we treat a "*" as a "." for IP addresses. */
				if (digit == '*')
					s[0] = '.';
				/* If it's a "#" and not at the beginning, we treat it as
				   an "enter" and don't want to append it */
				if (digit == '#') {
					gtk_button_clicked(GTK_BUTTON(callbttn));
					s[0] = '\0';
					return;
				}
			}
			if (s[0])
				gtk_entry_append_text(GTK_ENTRY(calladdr), s);
		}

		if (telco_get_mode() == TELCO_MODE_ASTERISK) {
			/* If we're using asterisk, send a dial plan request */
			if (tbd_pcid >= 0)
				pc_send_dpreq(SOURCE_GUI, tbd_pcid, gtk_entry_get_text(GTK_ENTRY(calladdr)));
		} else {
			/* Otherwise, compare it to our hard coded dial plan */
			if (telco_match_nbrc(gtk_entry_get_text(GTK_ENTRY(calladdr)))) {
				fprintf(stderr, "%s is a phone number!\n", gtk_entry_get_text(GTK_ENTRY(calladdr)));
				gtk_button_clicked(GTK_BUTTON(callbttn));
			}
		}
	}
}

static void callanswer_cb(GtkWidget *w, void *data)
{
	if (!callwin)
		show_call_window(1, 1);
	else
		gtk_button_clicked(GTK_BUTTON(callbttn));
}


static int peerevent(GtkWidget *w, GdkEvent *event, struct peer *p)
{
	int newx;
	int newy;
	switch(event->type) {
	case GDK_BUTTON_PRESS:
		p->xoff = event->button.x;
		p->yoff = event->button.y;
		p->moving = 1;
		break;
	case GDK_MOTION_NOTIFY:
		if (p->moving) {
			newx = p->box->allocation.x - p->xoff + event->motion.x;
			newy = p->box->allocation.y - p->yoff + event->motion.y;
			if (newx > fixed->allocation.width - 5)
				newx = fixed->allocation.width - 5;
			if (newy > fixed->allocation.height - 5)
				newy = fixed->allocation.height - 5;
			gtk_widget_set_uposition(p->box, 
				newx,
				newy);
		}
		break;
	case GDK_BUTTON_RELEASE:
		p->moving = 0;
		break;
	default:
	}
	return FALSE;
}

struct peer *peerwindow(int self, char *label)
{
	struct peer *p;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkAdjustment *a;
	GtkWidget *eventbox;
	p = g_new0(struct peer, 1);

	/* Start muted */
	p->mute = 1;

	p->laglock = 0;
	p->lagfd = -1;

	/* Build a vbox to hold each component set */
	p->outerbox = gtk_vbox_new(FALSE, 5);
	if (self) 
		gtk_box_pack_start(GTK_BOX(p->outerbox), build_self_win(), FALSE, FALSE, 10);
	else {
		p->label = gtk_label_new(label);
		gtk_label_set_line_wrap(GTK_LABEL(p->label), TRUE);
		gtk_box_pack_start(GTK_BOX(p->outerbox), p->label, FALSE, FALSE, 5);
		gtk_box_pack_start(GTK_BOX(p->outerbox), build_generic_win(p), FALSE, FALSE, 10);
	}
	/* Build a level meter and frame to put it in */
	p->level = gtk_progress_bar_new();
	
	
	/* Build the volume */
	a = (GtkAdjustment *)gtk_adjustment_new(DEFAULT_SUPPRESSION, 0.0, 100.0, 1.0, 1.0, 0.0);
	p->volume = gtk_hscale_new(a);
	p->pcid = -1;
	/* We want a separate event box so that moving the slider doesn't move the window */
	eventbox = gtk_event_box_new();
	gtk_container_add(GTK_CONTAINER(eventbox), p->volume);
	gtk_widget_set_usize(p->level, 100, 25);
	vbox = gtk_vbox_new(FALSE, 0);
	p->frame = gtk_frame_new("Level");
	gtk_frame_set_shadow_type(GTK_FRAME(p->frame), GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(vbox), p->level, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), eventbox, FALSE, FALSE, 0);

	
	if (!self) { /* We won't have lag with ourselves */
		p->lagmeter = gtk_label_new("Lag/Jitter: (calculating)");
		gtk_box_pack_start(GTK_BOX(p->outerbox), p->lagmeter, FALSE, FALSE, 0);
	}
	
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
	gtk_container_add(GTK_CONTAINER(p->frame), vbox);
	gtk_box_pack_start(GTK_BOX(p->outerbox), p->frame, FALSE, FALSE, 0);

	/* Add the dialpad */
	if (self)
		gtk_box_pack_start(GTK_BOX(p->outerbox), build_dialpad(), FALSE, FALSE, 0);

	/* Build a re-parentable box, and then put the stuff
	   in it */
	if (self) 
		p->box = gtk_frame_new(label);
	else {
		p->box = gtk_frame_new(NULL);
		gtk_frame_set_shadow_type(GTK_FRAME(p->box), GTK_SHADOW_OUT);
	}
	if (self) {
		hbox = gtk_hbox_new(FALSE, 5);
		callanswer = gtk_button_new_with_label("Call");
		gtk_signal_connect(GTK_OBJECT(callanswer), "clicked", GTK_SIGNAL_FUNC(callanswer_cb), p);
		gtk_widget_set_usize(callanswer, 80, 30);
		gtk_box_pack_start(GTK_BOX(hbox), gtk_event_box_new(), TRUE, TRUE, 5);
		gtk_box_pack_start(GTK_BOX(hbox), callanswer, FALSE, FALSE, 5);
		gtk_box_pack_start(GTK_BOX(hbox), gtk_event_box_new(), TRUE, TRUE, 5);
		gtk_box_pack_start(GTK_BOX(p->outerbox), hbox, FALSE, FALSE, 0);
	} else {
		hbox = gtk_hbox_new(FALSE, 5);
		p->hold = gtk_toggle_button_new_with_label("Hold");
		gtk_signal_connect(GTK_OBJECT(p->hold), "clicked", GTK_SIGNAL_FUNC(peerhold_cb), p);
		p->conf = gtk_toggle_button_new_with_label("Conf");
		gtk_signal_connect(GTK_OBJECT(p->conf), "clicked", GTK_SIGNAL_FUNC(conference_cb), p);
		p->direct = gtk_toggle_button_new_with_label("Direct");
		gtk_signal_connect(GTK_OBJECT(p->direct), "clicked", GTK_SIGNAL_FUNC(direct_cb), p);
		fake_set_active(p->direct, TRUE);
		gtk_widget_set_usize(p->conf, 50, 20);
		gtk_widget_set_usize(p->hold, 50, 20);
		gtk_widget_set_usize(p->direct, 50, 20);
		gtk_box_pack_start(GTK_BOX(hbox), p->hold, FALSE, FALSE, 5);
		gtk_box_pack_start(GTK_BOX(hbox), p->conf, FALSE, FALSE, 5);
		gtk_box_pack_start(GTK_BOX(hbox), p->direct, FALSE, FALSE, 5);
		gtk_box_pack_start(GTK_BOX(p->outerbox), hbox, FALSE, FALSE, 0);
		hbox = gtk_hbox_new(FALSE, 5);
		p->hangup = gtk_button_new_with_label("Hangup");
		gtk_signal_connect(GTK_OBJECT(p->hangup), "clicked", GTK_SIGNAL_FUNC(peerhangup_cb), p);
		p->answer = gtk_button_new_with_label("Answer");
		gtk_signal_connect(GTK_OBJECT(p->answer), "clicked", GTK_SIGNAL_FUNC(peeranswer_cb), p);
		gtk_widget_set_usize(p->answer, 70, 30);
		gtk_widget_set_usize(p->hangup, 70, 30);
		gtk_box_pack_start(GTK_BOX(hbox), p->answer, FALSE, FALSE, 5);
		gtk_box_pack_end(GTK_BOX(hbox), p->hangup, FALSE, FALSE, 5);
		gtk_box_pack_start(GTK_BOX(p->outerbox), hbox, FALSE, FALSE, 0);
	}
	p->status = gtk_statusbar_new();
	gtk_box_pack_start(GTK_BOX(p->outerbox), p->status, FALSE, FALSE, 0);

	gtk_container_set_border_width(GTK_CONTAINER(p->outerbox), 5);
	gtk_container_add(GTK_CONTAINER(p->box), p->outerbox);
	
	if (!self) {
		/* Lets put ourselves in an event box */
		eventbox = gtk_event_box_new();
		gtk_signal_connect(GTK_OBJECT(eventbox), "event", GTK_SIGNAL_FUNC(peerevent), p);
		gtk_container_add(GTK_CONTAINER(eventbox), p->box);
		p->frame2 = p->box;
		p->box = eventbox;
	}
	
	return p;	
}

static void send_url_home(void)
{
	interpret_call(DEFAULT_HOME_PAGE);
}

static void send_url_register(void)
{
	interpret_call(REGISTER_URL);
}

static void send_url_bugs(void)
{
	interpret_call(BUGS_URL);
}

static void send_url_support(void)
{
	interpret_call(SUPPORT_URL);
}

static void send_url_developers(void)
{
	interpret_call(DEVELOPERS_URL);
}

static void send_url_demo(void)
{
	interpret_call(INTRODUCTORY_CALL);
}

static void upgrade_init(void)
{
	upgrade_start(mw);
}

int build_gui(void)
{
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *frame;
	GtkItemFactory *gif;
	
	static GtkItemFactoryEntry menu_items[] = {
	{ "/_File", NULL, NULL, 0, "<Branch>" },
	{ "/File/New _Call", "<control>C", GTK_SIGNAL_FUNC(open_new_call) },
	{ "/File/_Open URL","<control>O", GTK_SIGNAL_FUNC(open_new_url) },
#ifdef ENABLE_UPGRADE
	{ "/File/_Upgrade","<control>U", GTK_SIGNAL_FUNC(upgrade_init) },
#endif
	{ "/File/", NULL, NULL, 0, "<Separator>" },
	{ "/File/_Quit", "<control>Q", GTK_SIGNAL_FUNC(quit) },
	{ "/_Gnophone.com", NULL, NULL, 0, "<Branch>" },
	{ "/_Gnophone.com/Home", "<control>H", GTK_SIGNAL_FUNC(send_url_home) },
	{ "/_Gnophone.com/Introduction", "<control>I", GTK_SIGNAL_FUNC(send_url_demo) },
	{ "/_Gnophone.com/Register", "<control>R", GTK_SIGNAL_FUNC(send_url_register) },
	{ "/_Gnophone.com/", NULL, NULL, 0, "<Separator>" },
	{ "/_Gnophone.com/Bug Reports", "<control>B", GTK_SIGNAL_FUNC(send_url_bugs) },
	{ "/_Gnophone.com/Support", "<control>S", GTK_SIGNAL_FUNC(send_url_support) },
	{ "/_Gnophone.com/Developers", "<control>D", GTK_SIGNAL_FUNC(send_url_developers) },
	{ "/_Preferences", NULL, NULL, 0, "<Branch>" },
	{ "/Preferences/_Audio...", "<control>A", GTK_SIGNAL_FUNC(audio_prefs) },
	{ "/Preferences/_Video...", "<control>V", GTK_SIGNAL_FUNC(image_callback) },
	{ "/Preferences/_Telephone...", "<control>T", GTK_SIGNAL_FUNC(telco_prefs) },
	{ "/_Help", NULL, NULL, 0, "<LastBranch>" },
	{ "/Help/_About Gnophone", NULL, GTK_SIGNAL_FUNC(show_about) },
	
	};
	int items = sizeof(menu_items) / sizeof(menu_items[0]);
	
	/* Build the top level window */
	mw = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_realize(mw);
	gtk_signal_connect(GTK_OBJECT(mw), "destroy", GTK_SIGNAL_FUNC(quit), NULL);
	gtk_window_set_title(GTK_WINDOW(mw), "Gnophone");

	/* Build our accelerator group */
	gag = gtk_accel_group_new();
	
	/* Build the menubar */
	gif = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", gag);
	gtk_item_factory_create_items (gif, items, menu_items, NULL);
  	menubar = gtk_item_factory_get_widget(gif, "<main>");
	
	/* Build the notebook itself */
	nb = gtk_notebook_new();

	/* Build the browser */
	pbrowser = build_browser();

	if (pbrowser) {
		/* Add the browser as a page */
		gtk_notebook_append_page(GTK_NOTEBOOK(nb), pbrowser->box, gtk_label_new("Browser"));
	}

	/* And the fixed area */
	frame = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
	
	fixed = gtk_fixed_new();
	fixedstatus = gtk_statusbar_new();
	vbox = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(frame), fixed);
	gtk_box_pack_start(GTK_BOX(vbox), frame, TRUE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX(vbox), fixedstatus, FALSE, FALSE, 5);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
	gtk_notebook_append_page(GTK_NOTEBOOK(nb), vbox, gtk_label_new("Calls"));

	/* No peers right now -- don't confuse the user */
	setpeerstatus(NULL, "You have no calls.");

	/* Build the first vbox and place the notebook in it */
	vbox = gtk_vbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(vbox), nb, TRUE, TRUE, 5);
	
	/* Build the hbox and place the vbox in it */
	hbox = gtk_hbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 5);

	/* Make ourselves and put us "in the box" */
	self = peerwindow(1, "You!");
	vbox = gtk_vbox_new(FALSE, 5);
	gtk_box_pack_end(GTK_BOX(vbox), self->box, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 5);

	/* Put the hbox in a vbox, with the menu above it */
	vbox = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), menubar, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

	gtk_container_add(GTK_CONTAINER(mw), vbox);
	gtk_widget_set_usize(mw, 700, 540);
	gtk_window_add_accel_group(GTK_WINDOW(mw), gag);
	if (pbrowser)
		gtk_widget_realize(pbrowser->html);
	gtk_widget_show_all(mw);
	mk_image(gwin, get_image_filename());
	return 0;
}

void open_new_call(GtkWidget *w)
{
	show_call_window(0, 0);
}

void open_new_url(GtkWidget *w)
{
	show_call_window(0, 0);
}

static float loudness(struct peer *p, int level)
{
	float sum=level;
	sum /= 32768; 
	if (sum > p->history)
		p->history = sum;
	else
		p->history = sum/4 + 3.0 * p->history/4.0;
	if (p->history > 1.0)
		p->history = 1.0;
	p->blah = p->history * 3.0;
	if (p->blah > 1.0)
		p->blah = 1.0;
	return p->blah;
}

#if 0
static int calc_ms()
{
	static struct timeval last = {0,0};
	struct timeval tv;
	int ms;
	gettimeofday(&tv, NULL);
	if (last.tv_sec)
		ms = (tv.tv_sec - last.tv_sec) * 1000 + (tv.tv_usec - last.tv_usec) / 1000;
	else
		ms = 0;
	last = tv;
	return ms;
}
#endif

static struct peer *find_peer(int pcid)
{
	struct peer *cur = peers;
	while(cur) {
		if (cur->pcid == pcid)
			return cur;
		cur = cur->next;
	}
	return NULL;
}

static void load_image(struct peer *p, void *data, int datalen)
{
	int fd;
	char filename[80];
	if (getenv("HOME")) {
		snprintf(filename, sizeof(filename), "%s/.gnophone/tmp", getenv("HOME"));
		mkdir(filename, 0775);
		snprintf(filename, sizeof(filename), "%s/.gnophone/tmp/tmpimg-XXXXXX", getenv("HOME"));
	} else
		snprintf(filename, sizeof(filename), "/tmp/tmpimg-XXXXXX");
	fd = mkstemp(filename);
	if (fd < 0) {
		fprintf(stderr, "Unable to create temporary file: %s\n", strerror(errno));
		return;
	} else {
		if (write(fd, data, datalen) != datalen) {
			fprintf(stderr, "Unable to write image: %s\n", strerror(errno));
			close(fd);
			return;
		}
		close(fd);
		mk_image(p->image, filename);
		unlink(filename);
	}
	if (nb)
		gtk_notebook_set_page(GTK_NOTEBOOK(nb), 1);
}

static char passwd[256];
static GtkWidget *passwin;
static GtkWidget *password;
static int retval;

static int killpasswd()
{
	gtk_grab_remove(passwin);
	passwin = NULL;
	return TRUE;
}

static void passwd_cancel()
{
	gtk_widget_destroy(passwin);
}

static void passwd_okay()
{
	strncpy(passwd, gtk_entry_get_text(GTK_ENTRY(password)), sizeof(passwd)-1);
	retval = 0;
	gtk_widget_destroy(passwin);
}

static void build_password(struct sockaddr_in sin, char *username)
{
	char buf[80];
	GtkWidget *table;
	GtkWidget *label;
	GtkWidget *button;
	snprintf(buf, sizeof(buf), "%s:%d", inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));
	table = gtk_table_new(5,2, FALSE);
	label = gtk_label_new("Password Required");
	gtk_table_attach(GTK_TABLE(table), label, 0, 2, 0, 1, GTK_EXPAND, GTK_EXPAND, 10, 20);

	label = gtk_label_new("Server: ");
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2, GTK_EXPAND, GTK_EXPAND, 10, 5);
	label = gtk_label_new(buf);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
	gtk_table_attach(GTK_TABLE(table), label, 1, 2, 1, 2, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 10, 5);

	label = gtk_label_new("Username: ");
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 2, 3, GTK_EXPAND, GTK_EXPAND, 10, 5);
	label = gtk_label_new(username);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
	gtk_table_attach(GTK_TABLE(table), label, 1, 2, 2, 3, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 10, 5);

	label = gtk_label_new("Password: ");
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 3, 4, GTK_EXPAND, GTK_EXPAND, 10, 5);
	password = gtk_entry_new();
	gtk_entry_set_visibility(GTK_ENTRY(password), FALSE);
	gtk_table_attach(GTK_TABLE(table), password, 1, 2, 3, 4, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 10, 5);
	gtk_signal_connect(GTK_OBJECT(password), "activate", GTK_SIGNAL_FUNC(passwd_okay), NULL);

	button = gtk_button_new_with_label("OK");
	gtk_widget_set_usize(button, 100, 30);
	gtk_table_attach(GTK_TABLE(table), button, 0, 1, 4, 5, GTK_EXPAND, GTK_EXPAND, 10, 5);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(passwd_okay), NULL);

	button = gtk_button_new_with_label("Cancel");
	gtk_widget_set_usize(button, 100, 30);
	gtk_table_attach(GTK_TABLE(table), button, 1, 2, 4, 5, GTK_EXPAND, GTK_EXPAND, 10, 5);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(passwd_cancel), NULL);

	gtk_container_set_border_width(GTK_CONTAINER(table), 10);

	passwin = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_container_add(GTK_CONTAINER(passwin), table);
	gtk_window_set_title(GTK_WINDOW(passwin), "Gnophone: Password Required");
	gtk_signal_connect(GTK_OBJECT(passwin), "destroy", GTK_SIGNAL_FUNC(killpasswd), NULL);
	gtk_widget_show_all(passwin);	
	gtk_widget_grab_focus(password);
}

static int get_pass(struct peer *p, char *username, char **pass)
{
	/* The OK button will set retval to the proper
	   return value if necessary */
	retval = -1;
	bzero(passwd, sizeof(passwd));
	gtk_widget_set_sensitive(mw, FALSE);
	build_password(p->addr, username);
	while(passwin && !gtk_main_iteration_do(TRUE));
	gtk_widget_set_sensitive(mw, TRUE);
	*pass = passwd;
	
	return retval;
}

static void handle_event(pc_event *e, struct peer *p)
{
	struct hostent *hp;
	struct in_addr in;
	char *pass, *user;
	char buf[256];
	switch(e->event) {

	case PC_EVENT_ADDR:
		memcpy(&p->addr, &e->e.addr.addr, sizeof(p->addr));
		break;
	case PC_EVENT_HANGUP:
		handle_hangup(p, "Call disconnected by peer");
		break;
	case PC_EVENT_REJECT:
		snprintf(buf, sizeof(buf), "Call Rejected: %s", strlen(e->e.reject.why) ? e->e.reject.why : "No reason");
		handle_hangup(p, buf);
		break;
	case PC_EVENT_ACCEPT:
		setpeerstatus(p, "Waiting for answer...");

		if ((video_options & OPT_VIDEO_SEND_IMAGE) && (video_options & OPT_VIDEO_SEND_IMAGE_PREANSWER))
		{
			struct stat s;
			char buf[1024];

			if (getenv("HOME"))
			{
				snprintf(buf, sizeof(buf), "%s/.gnophone/beforetheyanswer.jpg", getenv("HOME"));
				printf("%s\n", buf);

				if (!stat(buf, &s))
				{
					send_my_image(e->callno, buf);
					break;
				}
			}

			if (user_image_filename())
				send_my_image(e->callno, user_image_filename());	
		}

		break;
	case PC_EVENT_ANSWER:
		setpeerstatus(p, "Call connected.");
		audio_shutup();
		select_peer(p);
		p->mute = 0;

		if (user_image_filename() && (video_options && OPT_VIDEO_SEND_IMAGE))
			send_my_image(e->callno, user_image_filename());	
		
		break;
	case PC_EVENT_AUTHRQ:
		/* First check to see if this is our asterisk server */
		if (telco_get_mode() == TELCO_MODE_ASTERISK) {
			pass = telco_get_password();
			user = telco_get_username();
			if (pass && strlen(pass) && user && strlen(user)
			    && e->e.authrequest.username &&
				!strcasecmp(e->e.authrequest.username, user) &&
				telco_get_server()) {
				/* They have a username and password, and this is
				   the request username.  They even have
				   a username. Lets make sure that  */
				hp = gethostbyname(telco_get_server());
				if (hp) {
					memcpy(&in, hp->h_addr, sizeof(in));
					if (in.s_addr == p->addr.sin_addr.s_addr) {
						fprintf(stderr, "Found a match!\n");
						pc_auth_reply(SOURCE_GUI, p->pcid, telco_get_password(), e->e.authrequest.challenge, 
										e->e.authrequest.authmethods);
						break;
					} else {
						fprintf(stderr, "Close but no cigar\n");
					}
				} else {
					fprintf(stderr, "%s isn't a server\n", telco_get_server());
				}
			}
		}
		if (!get_pass(p, e->e.authrequest.username, &pass))  {
			fprintf(stderr, "Replying to authentication\n");
			pc_auth_reply(SOURCE_GUI, p->pcid, pass, e->e.authrequest.challenge, 
							e->e.authrequest.authmethods); 
		} else {
			pc_hangup(SOURCE_GUI, p->pcid, "I refuse to authenticate");
			handle_hangup(p, "Authentication Required");
		}
		break;
	case PC_EVENT_TRANSFER:
		/* If we're here, then we were just transfered. It's possible that our image was lost, so, let's 
		* send it again */

		if (user_image_filename() && (video_options && OPT_VIDEO_SEND_IMAGE))
			send_my_image(e->callno, user_image_filename());	

		break;

	case PC_EVENT_IMAGE:
		/* We get to cheat -- Use imlib to automatically load the image, regardless of
		   format, etc.  The down side is that we have to create a temporary file -- yuck! */

		if (!(video_options & OPT_VIDEO_IGNORE_IMAGE))
		{
			if (e->e.image.datalen) 
				load_image(p, e->e.image.data, e->e.image.datalen);
			else
				fprintf(stderr, "notice: empty image\n");
		}
		break;
	case PC_EVENT_URL:
		iaxloading = 1;
		if (e->e.url.url && strlen(e->e.url.url)) {
			interpret_call(e->e.url.url);
			urlpeer = p;
			gtk_notebook_set_page(GTK_NOTEBOOK(nb), 0);
		}
		break;
	case PC_EVENT_RINGA:
		html_setstatus(pbrowser, "Ringing...");
		setpeerstatus(p, "Ringing...");
		audio_ringing();
		break;
	case PC_EVENT_AUDIO:
		gtk_progress_bar_update(GTK_PROGRESS_BAR(p->level), loudness(p, e->e.audio.level));
		e->event = PC_EVENT_AUDIO_READY;
		pc_write_event(SOURCE_GUI, e);
		break;
	case PC_EVENT_LAGREP:
		/* Looks like we got a lag reply. */


		if (p->lagmeter) { /* Don't do anythin gif it doesnt exist */
			gchar tmp[64];

			g_snprintf(tmp, sizeof(tmp), "Lag/Jitter: %dms / %dms", e->e.lag.lag, e->e.lag.jitter);

			/* Set the label text */
			gtk_label_set_text(GTK_LABEL(p->lagmeter), tmp);
		}

		break;
	default:
		printf("Unknown event: %d\n", e->event);
	}
}

static void get_best_pos(int *x, int *y)
{
	/* XXX Fix me, I'm lousy! XXX */
	struct peer *p;
	int minx;
	int miny = 0;
	int newminy;
	int numpeers;

tryagain:
	p = peers;
	numpeers = 0;
	minx = 0;
	printf("Trying %d miny\n", miny);
	while(p) {
		numpeers++;
		if ((p->box->allocation.width  + p->box->allocation.x > minx) &&
			(p->box->allocation.height + p->box->allocation.y > miny)) {
			minx = p->box->allocation.width + p->box->allocation.x;
			printf("Increasing search to %d, %d\n", minx, miny);
		}

		p = p->next;
	}
	if (minx > fixed->allocation.width) {
		minx = 0;
		newminy = 99999;
		p = peers;
		while(p) {
			if ((p->box->allocation.height  + p->box->allocation.y > miny) &&
			    (p->box->allocation.height + p->box->allocation.y < newminy)) {
					/* Increase by the minimum we can */
				newminy = p->box->allocation.height + p->box->allocation.y;
			}
			p = p->next;
		}
		if (newminy < 99999) {
			miny = newminy;
			goto tryagain;
		}
	}
	if (miny >fixed->allocation.height) {
		minx = numpeers * 20;
		miny = numpeers * 20;
	}
	*x = minx;
	*y = miny;
	printf("Best x is %d, best y is %d\n", minx, miny);
}


void do_phonecore()
{
	pc_event e;
	struct peer * peer;
	char buf[80];
	int x, y;
	int res;
	float pos = 0.0;
	int uc = use_cam();

	res = pc_read_event(SOURCE_GUI, &e);
	if (res < 0) {
		fprintf(stderr, "Failed to read from phonecore...\n");
	} else {
		peer = find_peer(e.callno);
		if (peer) {
			handle_event(&e, peer);
		} else {
			if (e.callno == -1) {
				switch(e.event) {
				case PC_EVENT_ONHOOK:
					hangup();
					hide_call_window();
					break;
				case PC_EVENT_OFFHOOK:
					answer();
					break;
				case PC_EVENT_DIAL:
					digit_ready(e.e.dial.digit, 1, 0);
					break;
				case PC_EVENT_AUDIO:
					if (self)
						pos = loudness(self, e.e.audio.level);
					if (other_loudness) {
						other_loudness(e.e.audio.level);
					}
					if (self && self->level)
						gtk_progress_bar_update(GTK_PROGRESS_BAR(self->level), pos);
					e.event = PC_EVENT_AUDIO_READY;
					pc_write_event(SOURCE_GUI, &e);
					break;
				case PC_EVENT_DIAL_TBD:

					show_call_window(0, 0);

					tbd_pcid = pc_session_new();

					/* FIXME: Make this not be hardcoded */
					if (pc_call(SOURCE_GUI, tbd_pcid, "Robbie", "tormenta:5036/TBD@default", NULL))
						printf("DEBUG: TBD has been dialed.\n");
						
					break;
				case PC_EVENT_DPREP:
					/* Now, we should complete the call */
					launch_call(e.e.dprep.number, e.e.dprep.number, 1);

					/* And hide our call window */
					hide_call_window();

					break;

				default:
					fprintf(stderr, "Don't know what to do with oob event %d\n", e.event);
				}
				return;
			}
#if 0
			printf("Event %d for callno %d, registry = %d\n", e.event, e.callno, registry);
#endif			
			if (e.callno == registry) {
				registry = -1;
				if (e.event == PC_EVENT_REGREP) {
					if (e.e.regreply.status == PC_REG_SUCCESS) {
						regrefresh = e.e.regreply.refresh;

						if (strlen(e.e.regreply.callerid) > 0) {
							/* Parse the caller ID info, if need be */
							char tmp1[32];
							char tmp2[32];
							int i,j;

							i = 1; j = 0;
							do {
								tmp1[j++] = e.e.regreply.callerid[i++];
							} while (e.e.regreply.callerid[i] != '\"');

							tmp1[j] = 0; i++;

							if (e.e.regreply.callerid[i] == '0')
								do {
								} while (e.e.regreply.callerid[++i] == ' ');

							strcpy(tmp2, e.e.regreply.callerid + i);
							
							snprintf(buf, sizeof(buf), "Registered '%s' at '%s' as '%s'", telco_get_peer(), telco_get_server(), tmp2);
							gtk_frame_set_label(GTK_FRAME(self->box), tmp1);
						} else {
							/* If we didn't get any caller ID info, let's just leave things how they are :-D */
							snprintf(buf, sizeof(buf), "Registered '%s' at '%s'", telco_get_peer(), telco_get_server());
						}

						
					} else {
						regrefresh = 60;
						snprintf(buf, sizeof(buf), "Registration failed for '%s' at '%s'", telco_get_peer(), telco_get_server());
					}
					setpeerstatus(NULL, buf);
					html_setstatus(pbrowser, buf);
					schedule_registry();
					return;
				} else
					fprintf(stderr, "Huh? Got an event of type %d on the registry\n", e.event);
			}
			/* This must be a new connection */
			if (e.event != PC_EVENT_CONNECT) {
				fprintf(stderr, "Huh?  This is an event for a non-existant session %d?\n", e.callno);
				return;
			}
			sessions++;
			if (sessions >= MAX_SESSIONS) {
				setpeerstatus(NULL, "Missed a call");
				return;
			}
			if (strlen(e.e.connect.callerid) && strlen(e.e.connect.dnid)) {
				snprintf(buf, sizeof(buf), "Call from '%s' for '%s'", e.e.connect.callerid, 
							e.e.connect.dnid);
			} else if (strlen(e.e.connect.dnid)) {
				snprintf(buf, sizeof(buf), "Call for '%s'", e.e.connect.dnid);
			} else if (strlen(e.e.connect.callerid)) {
				snprintf(buf, sizeof(buf), "Call from '%s'", e.e.connect.callerid);
			} else {
				snprintf(buf, sizeof(buf), "Call from %s", inet_ntoa(e.e.connect.addr.sin_addr));
			}
			get_best_pos(&x, &y);
			peer = peerwindow(0, buf);
			peer->next = peers;
			peers = peer;
			peer->pcid = e.callno;
			memcpy(&peer->addr, &e.e.connect.addr, sizeof(peer->addr));
			gtk_widget_show_all(peer->box);

			if (!uc)
				gtk_widget_hide(peer->sendimage);

			gtk_fixed_put(GTK_FIXED(fixed), peer->box, x, y);
			gtk_notebook_set_page(GTK_NOTEBOOK(nb), 1);
		
			if (uc)
				stop_camera();

			while(gtk_events_pending())
				gtk_main_iteration_do(FALSE);

			if (uc)
				start_camera();

			/* Send an initial lag request */
			send_lag_request(peer);

			mk_image(peer->image, base_filename(DEFAULT_IMAGE));
			pc_accept(SOURCE_GUI, peer->pcid);
			pc_ring_announce(SOURCE_GUI, peer->pcid);
			most_recent_answer = peer;
			
			if (!current_peer)
				audio_ring();

			setpeerstatus(NULL, "Incoming call!");

			if ((video_options & OPT_VIDEO_SEND_IMAGE) && (video_options & OPT_VIDEO_SEND_IMAGE_PREANSWER))
			{
				struct stat s;
				char buf[1024];
				int didit = 0;

				if (getenv("HOME"))
				{
					snprintf(buf, sizeof(buf), "%s/.gnophone/beforeianswer.jpg", getenv("HOME"));
					printf("%s\n", buf);

					if (!stat(buf, &s))
					{
						send_my_image(peer->pcid, buf);
						didit = 1;
					}
				}

				if (!didit)
					if (user_image_filename())
						send_my_image(peer->pcid, user_image_filename());	
			}
		}
	}
}

static void iax_input(void *data, int fd, GdkInputCondition c)
{
	do_phonecore();
}



static int send_lag_request(gpointer data)
{
	struct peer *peer = (struct peer*)data;

	/* Lock the door */
	peer->laglock = 1;

	if (peer->lagfd < 0)
		peer->lagfd = gtk_timeout_add(LAG_TIMEOUT, send_lag_request, data);

	/* Send the lag request to the phonecore */	
	pc_send_lagreq(SOURCE_GUI, peer->pcid);

	/* Unlock the door */
	peer->laglock = 0;

	return TRUE;
}

int launch_call(char *destination, char *title, int tbd)
{
	struct peer *peer;
	char buf[80];
	char callerid[80];
	int x, y;
	int pcid;
	int uc = use_cam();


	/* If we're doing a tbd call, lets use call complete. */
	if (tbd) {
		if (pc_complete(SOURCE_GUI, tbd_pcid, destination)) {
			snprintf(buf, sizeof(buf), "Invalid destination: %s", destination);
			html_setstatus(pbrowser, buf);
			setpeerstatus(NULL, buf);
			return -1;
		}

		pcid = tbd_pcid;
		tbd_pcid = -1;

	} else {
		/* Otherwise, let's just make a new call */
		pcid = pc_session_new();
		if (gethostname(callerid, sizeof(callerid))) 
			strcpy(callerid, "Gnophone User");
		
		if (pc_call(SOURCE_GUI, pcid, callerid, destination, NULL)) {
			snprintf(buf, sizeof(buf), "Invalid destination: %s", destination);
			html_setstatus(pbrowser, buf);
			setpeerstatus(NULL, buf);
			return -1;
		}
	}

	snprintf(buf, sizeof(buf), "Call to %s", title);
	get_best_pos(&x, &y);
	peer = peerwindow(0, buf);
	peer->next = peers;
	peers = peer;
	peer->pcid = pcid;
	sessions++;
	gtk_widget_show_all(peer->box);

	if (!uc)
		gtk_widget_hide(peer->sendimage);
	
	gtk_fixed_put(GTK_FIXED(fixed), peer->box, x, y);
	gtk_notebook_set_page(GTK_NOTEBOOK(nb), 1);

	/* Send an initial lag request */
	send_lag_request(peer);

	if (uc)
		stop_camera();

	while(gtk_events_pending())
		gtk_main_iteration_do(FALSE);
	
	if (uc)
		start_camera();

	mk_image(peer->image, base_filename(DEFAULT_IMAGE));
	gtk_widget_set_sensitive(peer->answer, FALSE);
	setpeerstatus(NULL, "Attempting connection");
	select_peer(peer);

	return 0;
}

void iax_reg(void)
{
	char *peer = telco_get_peer();
	char *secret = telco_get_secret();
	char *server = telco_get_server();
	char buf[256];
	if (regid > -1) {
		gtk_timeout_remove(regid);
		regid =  -1;
		regrefresh = 60;
	}
	/* Schedule next udpate */
	schedule_registry();
	if (registry > -1) {
		/* Stop if the last registry hasn't finished yet */
		fprintf(stderr, "Registry still here...  Hanging it up\n");
		pc_hangup(SOURCE_GUI, registry, NULL);
	}
	if (peer && strlen(peer) && server && strlen(server)) {
		registry = pc_session_new();
		if (registry > -1) {
			pc_register(SOURCE_GUI, registry, server, peer, secret, 60);
			snprintf(buf, sizeof(buf), "Attempting to register '%s' at '%s'", peer, server);
		} else
			snprintf(buf, sizeof(buf), "Failed to register '%s' at '%s'", peer, server);
		html_setstatus(pbrowser, buf);
	}
}

int run_gui(void)
{
	gdk_input_add(pc_get_fd(), GDK_INPUT_READ, iax_input, NULL);
	iax_reg();
	gtk_main();
	return 0;
}

