/*
 * Gnophone: A client for the Asterisk PBX
 *
 * Copyright (C) 2000, Linux Support Services, Inc.
 *
 * Written by Mark Spencer
 *
 * Linux/UNIX version distributed under the terms of
 * the GNU General Public License
 *
 * html-mozilla.c Mozilla based browser back-end
 *
 */

#include <stdio.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>                                                   
#include <string.h>

#include <gtkmozembed.h>

#include "htmlengine.h"
#include "gnophone.h"

int cheap_js_hack = 0;

/* Working around mozilla not sending us the event when you click on a non-standard URL */

/*
#define VERY_TWISTED_LOGIC
*/

/*
 * Delay the load URL until we're back in the main
 * thread and not in mozilla
 */
#define SOMEWHAT_TWISTED_LOGIC

#ifdef VERY_TWISTED_LOGIC
#include <sys/time.h>
#endif /* VERY_TWISTED_LOGIC */

#define NOTE 	printf("%s\n", __PRETTY_FUNCTION__)

struct html_pvt {
#ifdef VERY_TWISTED_LOGIC
	GtkWidget *eventbox;
#endif	/* VERY_TWISTED_LOGIC */
#ifdef SOMEWHAT_TWISTED_LOGIC
	char *uri;
	int id;
#endif /* SOMEWHAT_TWISTED_LOGIC */
	GtkWidget *embed;
};

static void location_changed_cb(GtkMozEmbed *embed, struct browser *b)
{
	html_display_new_location(b, gtk_moz_embed_get_location(embed));
}

static void title_changed_cb(GtkMozEmbed *embed, struct browser *b)
{
	html_display_new_title(b, gtk_moz_embed_get_title(embed));
}

static void load_started_cb(GtkMozEmbed *embed, struct browser *b)
{
	html_display_load_started(b);
}

static void load_finished_cb(GtkMozEmbed *embed, struct browser *b)
{
	html_display_load_finished(b);
}

static int moz_can_go_back(struct browser *b)
{
	return gtk_moz_embed_can_go_back(GTK_MOZ_EMBED(b->pvt->embed));
}

static int moz_can_go_forward(struct browser *b) 
{
	return gtk_moz_embed_can_go_forward(GTK_MOZ_EMBED(b->pvt->embed));
}

static void net_status_change_cb(GtkMozEmbed *embed, int flags, struct browser *b)
{
#if 0
	if (flags & GTK_MOZ_EMBED_FLAG_WIN_START) {
		printf("[Win_Start]\n");
		flags &= ~GTK_MOZ_EMBED_FLAG_WIN_START;
	}
	if (flags & GTK_MOZ_EMBED_FLAG_WIN_STOP) {
		printf("[Win_stop]\n");
		flags &= ~GTK_MOZ_EMBED_FLAG_WIN_STOP;
	}
	switch(flags) {
	case GTK_MOZ_EMBED_FLAG_NET_START:
		html_setstatus(b, "Loading page...");
		break;
	case GTK_MOZ_EMBED_FLAG_NET_STOP:
		html_setstatus(b, "Completed.");
		break;
	case GTK_MOZ_EMBED_FLAG_NET_DNS:
		html_setstatus(b, "Looking up host...");
		break;
	case GTK_MOZ_EMBED_FLAG_NET_CONNECTING:
		html_setstatus(b, "Connecting to host...");
		break;
	case GTK_MOZ_EMBED_FLAG_NET_REDIRECTING:
		html_setstatus(b, "Redirecting...");
		break;
	case GTK_MOZ_EMBED_FLAG_NET_FAILEDDNS:
		html_setstatus(b, "Host not found.");
		break;
	case GTK_MOZ_EMBED_FLAG_NET_FAILEDCONNECT:
		html_setstatus(b, "Connection to host failed.");
		break;
	case GTK_MOZ_EMBED_FLAG_NET_FAILEDTRANSFER:
		html_setstatus(b, "Page is empty");
		break;
	case GTK_MOZ_EMBED_FLAG_NET_USERCANCELLED:
		html_setstatus(b, "Connection cancelled");
		break;
	default:
		printf("Unknown flags: %d\n", flags);
		break;
	}
#endif
	printf("Net-status: ");
	if (flags & GTK_MOZ_EMBED_FLAG_START) 
		printf("[start] ");
	if (flags & GTK_MOZ_EMBED_FLAG_REDIRECTING) 
		printf("[redirecting] ");
	if (flags & GTK_MOZ_EMBED_FLAG_NEGOTIATING)
		printf("[negotiating] ");
	if (flags & GTK_MOZ_EMBED_FLAG_STOP)
		printf("[stop] ");
	if (flags & GTK_MOZ_EMBED_FLAG_IS_REQUEST)
		printf("[is request] ");
	if (flags & GTK_MOZ_EMBED_FLAG_IS_DOCUMENT)
		printf("[is document] ");
	if (flags & GTK_MOZ_EMBED_FLAG_IS_NETWORK)
		printf("[is network] ");
	if (flags & GTK_MOZ_EMBED_FLAG_IS_WINDOW)
		printf("[is window] ");
	printf("\n");
}

static void progress_change_cb(GtkMozEmbed *embed, int cur, int max, struct browser *b)
{
	html_update_progress(b, cur, max);
}

#ifdef VERY_TWISTED_LOGIC

/* 
 * Since mozilla stupidly doesn't tell us when you click on a non-standard
 * URL, we have to go through a lot of hoops (or try to write a protocol
 * handler of our own) to know.  This is done rather more verbosely than
 * necessary in order to understand how VTL works
 */

#define VTL_NONE 0

#define VTL_FAST_ENOUGH 5

#define VTL_EMPTY_LINK_MESSAGE 1
#define VTL_EVENT_LEAVE 2
#define VTL_EVENT_ENTER 3

static char *lastlink = NULL;
static struct timeval vtl_tv;
static int vtl_state = VTL_NONE;
static void reset_timer() 
{
	gettimeofday(&vtl_tv, NULL);
}

static int vtl_fast_enough()
{
	int ms;
	struct timeval tv;
	gettimeofday(&tv, NULL);
	ms = (tv.tv_sec - vtl_tv.tv_sec) * 1000 + 
		(tv.tv_usec - vtl_tv.tv_usec) / 1000;
	if (ms <= VTL_FAST_ENOUGH)
		return 1;
	return 0; 
}

static void very_twisted_logic_handle_link(char *s)
{
	if (strlen(s)) {
		if (lastlink)
			free(lastlink);
		lastlink = strdup(s);
	} else {
		vtl_state = VTL_EMPTY_LINK_MESSAGE;
		reset_timer();
	}
		
}

#endif /* VERY_TWISTED_LOGIC */

static void link_message_cb(GtkMozEmbed *embed, struct browser *b)
{
	if (cheap_js_hack)
		return;

	html_display_link(b, gtk_moz_embed_get_link_message(embed));
#ifdef VERY_TWISTED_LOGIC
	very_twisted_logic_handle_link(gtk_moz_embed_get_link_message(GTK_MOZ_EMBED(b->pvt->embed)));
#endif
}

static void js_status_cb(GtkMozEmbed *embed, struct browser *b)
{
	if (cheap_js_hack)
		cheap_js_hack = 0;
	else
		cheap_js_hack = 1;

	html_display_link(b, gtk_moz_embed_get_js_status(embed));
#ifdef VERY_TWISTED_LOGIC
	very_twisted_logic_handle_link(gtk_moz_embed_get_js_status(GTK_MOZ_EMBED(b->pvt->embed)));
#endif

}

static void chrome_visibility_cb(GtkMozEmbed *embed, gboolean visibility, GtkWidget *w)
{

	NOTE;

	if (!visibility)
		gtk_widget_hide(w);
	else
		gtk_widget_show_all(w);

}

static void size_to_cb(GtkMozEmbed *embed, gint width, gint height)
{
	NOTE;

	gtk_widget_set_usize(GTK_WIDGET(embed), width, height);

}

static void title_cb(GtkMozEmbed *embed, GtkWidget *w)
{
	char *title;

	title = gtk_moz_embed_get_title(embed);

	if (title)
	{
		if (strlen(title))
			gtk_window_set_title(GTK_WINDOW(w), title);
		free(title);
	}

} 

static void destroy_browser_cb(GtkMozEmbed *embed, GtkWidget *w)
{
	NOTE;

	gtk_widget_destroy(w);

}

static void new_window_cb(GtkMozEmbed *embed, GtkMozEmbed **retval, int chromemask, struct browser *b)
{
	NOTE;

	if (chromemask & GTK_MOZ_EMBED_FLAG_OPENASCHROME)
	{
		GtkWidget *w;
		GtkMozEmbed *newmoz;

		w = gtk_window_new(GTK_WINDOW_TOPLEVEL);

		newmoz = GTK_MOZ_EMBED(gtk_moz_embed_new());

		gtk_container_add(GTK_CONTAINER(w), GTK_WIDGET(newmoz));


		gtk_signal_connect(GTK_OBJECT(newmoz), "visibility",
				GTK_SIGNAL_FUNC(chrome_visibility_cb), w);

		gtk_signal_connect(GTK_OBJECT(newmoz), "size_to",
				GTK_SIGNAL_FUNC(size_to_cb), NULL);

		gtk_signal_connect(GTK_OBJECT(newmoz), "title",
				GTK_SIGNAL_FUNC(title_cb), w);


		gtk_signal_connect(GTK_OBJECT(newmoz), "destroy_browser",
				GTK_SIGNAL_FUNC(destroy_browser_cb), w);

		gtk_widget_realize(w);
		gtk_widget_realize(GTK_WIDGET(newmoz));
	
		*retval = newmoz;
	}

}

static void visibility_cb(GtkMozEmbed *embed, gboolean visibility, struct browser *b)
{
}

static void destroy_brsr_cb(GtkMozEmbed *embed, struct browser *b)
{
	NOTE;
}

static void load_if_needed(char *uri, struct browser *b)
{
	if (strncasecmp(uri, "http:", 5) &&
	    strncasecmp(uri, "https:", 6) &&
	    strncasecmp(uri, "ftp:", 4)) {
			fprintf(stderr, "Need help loading '%s'\n", uri);
			interpret_call_strict(uri);
	}
}

#ifdef SOMEWHAT_TWISTED_LOGIC
static int delayed_openuri(void *vb)
{
	struct browser *b = vb;
	load_if_needed(b->pvt->uri, b);
	free(b->pvt->uri);
	b->pvt->uri = NULL;
	b->pvt->id = 0;
	return FALSE;
}
#endif

static int open_uri_cb(GtkMozEmbed *embed, char *uri, struct browser *b)
{
	NOTE;
	printf("Wants to access %s\n", uri);
#ifdef SOMEWHAT_TWISTED_LOGIC
	if (b->pvt->uri)
		free(b->pvt->uri);
	b->pvt->uri = g_strdup(uri);
	if (b->pvt->id > 0)
		gtk_timeout_remove(b->pvt->id);
	b->pvt->id = gtk_timeout_add(10, delayed_openuri, b);
#else
	load_if_needed(uri, b);
#endif
	return FALSE;
}


static void signal_setup(GtkWidget *html, struct browser *b)
{
	gtk_signal_connect(GTK_OBJECT(html), "location", GTK_SIGNAL_FUNC(location_changed_cb), b);
	gtk_signal_connect(GTK_OBJECT(html), "title", GTK_SIGNAL_FUNC(title_changed_cb), b);
	gtk_signal_connect(GTK_OBJECT(html), "net_start", GTK_SIGNAL_FUNC(load_started_cb), b);
	gtk_signal_connect(GTK_OBJECT(html), "net_stop", GTK_SIGNAL_FUNC(load_finished_cb), b);
	gtk_signal_connect(GTK_OBJECT(html), "net_state", GTK_SIGNAL_FUNC(net_status_change_cb), b);
	gtk_signal_connect(GTK_OBJECT(html), "progress", GTK_SIGNAL_FUNC(progress_change_cb), b);
	gtk_signal_connect(GTK_OBJECT(html), "link_message", GTK_SIGNAL_FUNC(link_message_cb), b);
	gtk_signal_connect(GTK_OBJECT(html), "js_status", GTK_SIGNAL_FUNC(js_status_cb), b);
	gtk_signal_connect(GTK_OBJECT(html), "new_window", GTK_SIGNAL_FUNC(new_window_cb), b);
	gtk_signal_connect(GTK_OBJECT(html), "visibility", GTK_SIGNAL_FUNC(visibility_cb), b);
	gtk_signal_connect(GTK_OBJECT(html), "destroy_browser", GTK_SIGNAL_FUNC(destroy_brsr_cb), b);
	gtk_signal_connect(GTK_OBJECT(html), "open_uri", GTK_SIGNAL_FUNC(open_uri_cb), b);
#if 0
	gtk_signal_connect(GTK_OBJECT(html), "destroy", GTK_SIGNAL_FUNC(destroy_cb), b);
#endif
}

static void moz_go_back(struct browser *b)
{
	gtk_moz_embed_go_back(GTK_MOZ_EMBED(b->pvt->embed));
}

static void moz_go_forward(struct browser *b)
{
	gtk_moz_embed_go_forward(GTK_MOZ_EMBED(b->pvt->embed));
}

static void moz_reload(struct browser *b, int force)
{
	if (force) 
		gtk_moz_embed_reload(GTK_MOZ_EMBED(b->pvt->embed), GTK_MOZ_EMBED_FLAG_RELOADBYPASSPROXYANDCACHE);
	else
		gtk_moz_embed_reload(GTK_MOZ_EMBED(b->pvt->embed), 0);
}

static void moz_loadurl(struct browser *b, char *uri)
{
	gtk_moz_embed_load_url(GTK_MOZ_EMBED(b->pvt->embed), uri);
}

static void moz_stop(struct browser *b)
{
	gtk_moz_embed_stop_load(GTK_MOZ_EMBED(b->pvt->embed));
}

#ifdef VERY_TWISTED_LOGIC
static int eventbox_cb(GtkEventBox *w, GdkEvent *e, struct browser *b)
{
#if 0
	printf("Event: %d, state: %d\n", e->type, vtl_state);
#endif	
	switch(e->type) {
	case GDK_LEAVE_NOTIFY:
		if (vtl_state == VTL_EMPTY_LINK_MESSAGE)	
			vtl_state = VTL_EVENT_LEAVE;
		break;
	case GDK_ENTER_NOTIFY:
		if (vtl_state == VTL_EVENT_LEAVE) {
			vtl_state = VTL_EVENT_ENTER;
			if (vtl_fast_enough()) {
				fprintf(stderr, "You clicked on %s\n", lastlink);
				load_if_needed(lastlink, b);
			}
#if 0
			 else {
				fprintf(stderr, "Wasn't fast enough for %s\n", lastlink);
			}
#endif
		}
		break;
	default:
		break;
	}
	return FALSE;
}
#endif

GtkWidget *moz_new_engine(struct browser *b)
{
	struct html_pvt *mp;
	mp = g_new0(struct html_pvt, 1);
	mp->embed = gtk_moz_embed_new();
	b->pvt = mp;
	
	/* Setup Signals on engine */
	signal_setup(b->pvt->embed, b);
#ifdef VERY_TWISTED_LOGIC
	/* XXX Mozilla work-around, detect non-standard URL's by
	   an event box and some crazy logic XXX */
	b->pvt->eventbox = gtk_event_box_new();
	gtk_signal_connect(GTK_OBJECT(b->pvt->eventbox), "event", GTK_SIGNAL_FUNC(eventbox_cb), b);
	gtk_container_add(GTK_CONTAINER(b->pvt->eventbox), b->pvt->embed);
	return b->pvt->eventbox;
#else
	return b->pvt->embed;
#endif

}

static struct html_engine mozilla;

char *key() 
{
	return KEY;
}

char *name()
{
	return "mozilla";
}

int init()
{
	char *profile_path;
	
	if (!getenv("MOZILLA_FIVE_HOME"))
		return -1;

	/* We want to initlize our profile if we want to use SSL stuff */
	profile_path = g_strdup_printf("%s/%s", getenv("HOME"), ".gnophone");
	gtk_moz_embed_set_profile_path(profile_path, "gnophone");
	gtk_moz_embed_set_comp_path(getenv("MOZILLA_FIVE_HOME"));

	strcpy(mozilla.name, "Mozilla/5.0");
	mozilla.new_engine = moz_new_engine;
	mozilla.can_go_back = moz_can_go_back;
	mozilla.can_go_forward = moz_can_go_forward;
	mozilla.go_back = moz_go_back;
	mozilla.go_forward = moz_go_forward;
	mozilla.reload = moz_reload;
	mozilla.loadurl = moz_loadurl;
	mozilla.stop = moz_stop;
	mozilla.next = NULL;
	html_register_engine(&mozilla);

	return 0;
}

